# Copyright (C) 2014-2020 The BET Development Team

r"""
This subpackage provides methods for approximating gradients of
QoI maps and choosing optimal QoIs to use in the inverse problem.

* :mod:`~bet.sensitivity.gradients` provides methods for approximating gradients of QoI maps.
* :mod:`~bet.sensitivity.chooseQoIs` provides methods for choosing optimal QoIs to use in the inverse problem.
"""

__all__ = ['gradients', 'chooseQoIs']
