# Copyright (C) 2014-2020 The BET Development Team

r"""
This subpackage provides classes and methods for calculating the
probability measure :math:`P_{\Lambda}`.

* :mod:`~bet.calculateP.calculateP` provides methods for approximating probability densities in the measure-based approach.
* :mod:`~bet.calculateP.simpleFunP` provides methods for creating simple function approximations of probability densities for the measure-based approach.
* :mod:`~bet.calculateP.calculateR` provides methods for density-based approach.
* :mod:`~bet.calculateP.calculateError` provides methods for approximating numerical and sampling errors.
"""
__all__ = ['calculateP', 'simpleFunP', 'calculateError', 'calculateR']
