# Copyright (C) 2014-2020 The BET Development Team

"""
This package contains all of the tests for :program:`BET`. The package
structure mirrors the ``bet`` package structure.
"""
__all__ = ['test_calculateP', 'test_postProcess', 'test_sampling',
           'test_sensitivity', 'test_util', 'test_Comm', 'test_sample',
           'test_surrogates', 'problem_setups']
