# Copyright (C) 2014-2020 The BET Development Team

r"""
This subpackage contains

* :class:`~bet.postProcess.plotP` plots :math:`P` and/or volumes (:math:`\mu`) of Voronoi cells
* :class:`~bet.postProcess.plotDomains` plots the data domain :math:`\mathcal{D}` in 2D
* :class:`~bet.postProcess.postTools` has tools for postprocessing
* :class:`~bet.postProcess.compareP` has tools for comparing measures
"""
__all__ = ['plotP', 'plotDomains', 'postTools', 'plotVoronoi', 'compareP']
