# Copyright (C) 2014-2020 The BET Development Team

"""
This package contains all of the tests for :program:`BET`. The package
structure mirrors the ``bet`` package structure.
"""
__all__ = ['test_calculateP', 'test_simpleFunP',
           'test_calculateError', 'test_calculateR']
