# Copyright (C) 2014-2020 The BET Development Team

"""
This subpackage contains the test modules for the sampling subpackage.
"""
__all__ = ['test_basicSampling',
           'test_Lp_generalized_samples', 'test_useLUQ']
